import unittest

import numpy as np
import pandas as pd

from lightwood.helpers.ts import Differencer


class TestTSDifferencer(unittest.TestCase):
    def test_numerical(self):
        D = Differencer()
        D.fit(np.array([1, 2, 3]))

        reconstructed = D.inverse_transform(D.transform(np.array([4, 5, 6])))
        target = pd.Series([3, 4, 5])
        self.assertTrue(reconstructed.all() == target.all())

        reconstructed = D.inverse_transform(D.transform(np.array([4, 5, 6])), init=4)
        target = pd.Series([4, 5, 6])
        self.assertTrue(reconstructed.all() == target.all())



from typing import Dict, Union

import numpy as np
import pandas as pd
from hyperopt import hp
import neuralforecast as nf
from neuralforecast.models.mqnhits.mqnhits import MQNHITS

from lightwood.helpers.log import log
from lightwood.mixer.base import BaseMixer
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs


class NHitsMixer(BaseMixer):
    horizon: int
    target: str
    supports_proba: bool
    model_path: str
    hyperparam_search: bool
    default_config: dict

    def __init__(
            self,
            stop_after: float,
            target: str,
            horizon: int,
            ts_analysis: Dict,
            pretrained: bool = False
    ):
        """
        Wrapper around a MQN-HITS deep learning model.
        
        :param stop_after: time budget in seconds.
        :param target: column to forecast.
        :param horizon: length of forecasted horizon.
        :param ts_analysis: dictionary with miscellaneous time series info, as generated by 'lightwood.data.timeseries_analyzer'.
        """  # noqa
        super().__init__(stop_after)
        self.stable = True
        self.prepared = False
        self.supports_proba = False
        self.target = target
        self.horizon = horizon
        self.ts_analysis = ts_analysis
        self.grouped_by = ['__default'] if not ts_analysis['tss'].group_by else ts_analysis['tss'].group_by

        self.pretrained = pretrained  # finetuning?
        self.base_url = 'https://nixtla-public.s3.amazonaws.com/transfer/pretrained_models/'
        self.freq_to_model = {
            'Y': 'yearly',
            'Q': 'monthly',
            'M': 'monthly',
            'W': 'daily',
            'D': 'daily',
            'H': 'hourly',
            'T': 'hourly',  # consider using another pre-trained model once available
            'S': 'hourly'  # consider using another pre-trained model once available
        }
        self.model_names = {
            'hourly': 'nhits_m4_hourly.ckpt',  # hourly (non-tiny)
            'daily': 'nhits_m4_daily.ckpt',   # daily
            'monthly': 'nhits_m4_monthly.ckpt',  # monthly
            'yearly': 'nhits_m4_yearly.ckpt',  # yearly
        }
        self.model_name = None
        self.model = None

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Fits the N-HITS model.
        """  # noqa
        log.info('Started fitting N-HITS forecasting model')

        cat_ds = ConcatedEncodedDs([train_data, dev_data])
        oby_col = self.ts_analysis["tss"].order_by
        df = cat_ds.data_frame.sort_values(by=f'__mdb_original_{oby_col}')

        # 2. adapt data into the expected DFs
        Y_df = self._make_initial_df(df)

        # set val-test cutoff
        n_time = len(df[f'__mdb_original_{oby_col}'].unique())
        n_ts_val = int(.1 * n_time)
        n_ts_test = int(.1 * n_time)

        # train the model
        n_time_out = self.horizon
        if self.pretrained:
            self.model_name = self.model_names.get(self.freq_to_model[self.ts_analysis['sample_freqs']['__default']],
                                                   None)
            self.model_name = self.model_names['hourly'] if self.model_name is None else self.model_name
            ckpt_url = self.base_url + self.model_name
            self.model = MQNHITS.load_from_checkpoint(ckpt_url)
            if self.horizon > self.model.hparams.n_time_out:
                self.pretrained = False

        if not self.pretrained:
            self.model = nf.auto.MQNHITS(horizon=n_time_out)
            self.model.space['max_steps'] = hp.choice('max_steps', [1e4])
            self.model.space['max_epochs'] = hp.choice('max_epochs', [50])
            self.model.space['n_time_in'] = hp.choice('n_time_in', [self.ts_analysis['tss'].window])
            self.model.space['n_time_out'] = hp.choice('n_time_out', [self.horizon])
            self.model.space['n_x_hidden'] = hp.choice('n_x_hidden', [0])
            self.model.space['n_s_hidden'] = hp.choice('n_s_hidden', [0])
            self.model.space['frequency'] = hp.choice('frequency', [self.ts_analysis['sample_freqs']['__default']])
            self.model.space['random_seed'] = hp.choice('random_seed', [42])
            self.model.fit(Y_df=Y_df,
                           X_df=None,       # Exogenous variables
                           S_df=None,       # Static variables
                           hyperopt_steps=5,
                           n_ts_val=n_ts_val,
                           n_ts_test=n_ts_test,
                           results_dir='./results/autonhits',
                           save_trials=False,
                           loss_function_val=nf.losses.numpy.mqloss,
                           loss_functions_test={'MQ': nf.losses.numpy.mqloss},
                           return_test_forecast=False,
                           verbose=True)

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Due to how lightwood implements the `update` procedure, expected inputs for this method are:
        
        :param dev_data: original `test` split (used to validate and select model if ensemble is `BestOf`).
        :param train_data: concatenated original `train` and `dev` splits.
        """  # noqa
        self.hyperparam_search = False
        self.fit(dev_data, train_data)
        self.prepared = True

    def __call__(self, ds: Union[EncodedDs, ConcatedEncodedDs],
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Calls the mixer to emit forecasts.
        
        NOTE: in the future we may support predicting every single row efficiently. For now, this mixer
        replicates the neuralforecast library behavior and returns a forecast strictly for the next `tss.horizon`
        timesteps after the end of the input dataframe.
        """  # noqa
        if args.predict_proba:
            log.warning('This mixer does not output probability estimates')

        length = sum(ds.encoded_ds_lenghts) if isinstance(ds, ConcatedEncodedDs) else len(ds)
        ydf = pd.DataFrame(0,  # zero-filled
                           index=np.arange(length),
                           columns=['prediction', 'lower', 'upper'],
                           dtype=object)

        input_df = self._make_initial_df(ds.data_frame).reset_index()
        ydf['index'] = input_df['index']

        pred_cols = ['y_5', 'y_50', 'y_95']
        target_cols = ['lower', 'prediction', 'upper']
        for target_col in target_cols:
            ydf[target_col] = [[0 for _ in range(self.horizon)] for _ in range(len(ydf))]  # zero-filled arrays

        group_ends = []
        for group in input_df['unique_id'].unique():
            group_ends.append(input_df[input_df['unique_id'] == group]['index'].iloc[-1])
        fcst = self.model.forecast(Y_df=input_df)

        for gidx, group in zip(group_ends, input_df['unique_id'].unique()):
            for pred_col, target_col in zip(pred_cols, target_cols):
                group_preds = fcst[fcst['unique_id'] == group][pred_col].tolist()[:self.horizon]
                idx = ydf[ydf['index'] == gidx].index[0]
                ydf.at[idx, target_col] = group_preds

        ydf['confidence'] = 0.9
        return ydf

    def _make_initial_df(self, df):
        oby_col = self.ts_analysis["tss"].order_by
        Y_df = pd.DataFrame()
        Y_df['y'] = df[self.target]
        Y_df['ds'] = pd.to_datetime(df[f'__mdb_original_{oby_col}'], unit='s')
        if self.grouped_by != ['__default']:
            Y_df['unique_id'] = df[self.grouped_by].apply(lambda x: ','.join([elt for elt in x]), axis=1)
        else:
            Y_df['unique_id'] = ''
        return Y_df



from typing import List, Tuple, Union, Dict

import numpy as np
import pandas as pd
from statsmodels.tsa.stattools import pacf


def get_ts_groups(df: pd.DataFrame, tss) -> list:
    group_combinations = ['__default']
    if tss.group_by:
        groups = [tuple([g]) if not isinstance(g, tuple) else g
                  for g in list(df.groupby(by=tss.group_by).groups.keys())]
        group_combinations.extend(groups)
    return group_combinations


def get_group_matches(
        data: Union[pd.Series, pd.DataFrame],
        combination: tuple,
        group_columns: List[str]
) -> Tuple[list, pd.DataFrame]:
    """Given a particular group combination, return the data subset that belongs to it."""

    if type(data) == pd.Series:
        data = pd.DataFrame(data)
    elif type(data) != pd.DataFrame:
        raise Exception(f"Wrong data type {type(data)}, must be pandas.DataFrame or pd.Series")

    if combination == '__default':
        return list(data.index), data
    else:
        subset = data
        for val, col in zip(combination, group_columns):
            subset = subset[subset[col] == val]
        if len(subset) > 0:
            return list(subset.index), subset
        else:
            return [], pd.DataFrame()


def get_delta(
        df: pd.DataFrame,
        dtype_dict: dict,
        group_combinations: list,
        tss
) -> Tuple[Dict, Dict, Dict]:
    """
    Infer the sampling interval of each time series, by picking the most popular time interval observed in the training data.

    :param df: Dataframe with time series data.
    :param group_combinations: all tuples with distinct values for `TimeseriesSettings.group_by` columns, defining all available time series.
    :param tss: timeseries settings

    :return:
    Dictionary with group combination tuples as keys. Values are dictionaries with the inferred delta for each series.
    """  # noqa
    df = df.copy()
    original_col = f'__mdb_original_{tss.order_by}'
    order_col = original_col if original_col in df.columns else tss.order_by
    deltas = {"__default": df[order_col].astype(float).rolling(window=2).apply(np.diff).value_counts().index[0]}
    freq, period = detect_freq_period(deltas["__default"], tss)
    periods = {"__default": period}
    freqs = {"__default": freq}

    if tss.group_by:
        for group in group_combinations:
            if group != "__default":
                _, subset = get_group_matches(df, group, tss.group_by)
                if subset.shape[0] > 1:
                    deltas[group] = subset[order_col].rolling(window=2).apply(np.diff).value_counts().index[0]
                    freq, period = detect_freq_period(deltas[group], tss)
                    periods[group] = period
                    freqs[group] = freq
                else:
                    deltas[group] = 1.0
                    periods[group] = 1
                    freqs[group] = 'S'

    return deltas, periods, freqs


def get_inferred_timestamps(df: pd.DataFrame, col: str, deltas: dict, tss) -> pd.DataFrame:
    horizon = tss.horizon
    if tss.group_by:
        gby = [f'group_{g}' for g in tss.group_by]

    for (idx, row) in df.iterrows():
        last = [r for r in row[f'order_{col}'] if r == r][-1]  # filter out nans (safeguard; it shouldn't happen anyway)

        if tss.group_by:
            try:
                series_delta = deltas[tuple(row[gby].tolist())]
            except KeyError:
                series_delta = deltas['__default']
        else:
            series_delta = deltas['__default']
        timestamps = [last + t * series_delta for t in range(horizon)]

        if tss.horizon == 1:
            timestamps = timestamps[0]  # preserves original input format if horizon == 1

        df[f'order_{col}'].iloc[idx] = timestamps
    return df[f'order_{col}']


def add_tn_num_conf_bounds(data: pd.DataFrame, tss_args):
    """
    Add confidence (and bounds if applicable) to t+n predictions, for n>1
    TODO: active research question: how to guarantee 1-e coverage for t+n, n>1
    For now, (conservatively) increases width by the confidence times the log of the time step (and a scaling factor).
    """
    for col in ['confidence', 'lower', 'upper']:
        data[col] = data[col].astype(object)

    for idx, row in data.iterrows():
        error_increase = [row['confidence']] + \
                         [row['confidence'] * np.log(np.e + t / 2)  # offset by e so that y intercept is 1
                          for t in range(1, tss_args.horizon)]
        data['confidence'].iloc[idx] = [row['confidence'] for _ in range(tss_args.horizon)]

        preds = row['prediction']
        width = row['upper'] - row['lower']
        data['lower'].iloc[idx] = [pred - (width / 2) * modifier for pred, modifier in zip(preds, error_increase)]
        data['upper'].iloc[idx] = [pred + (width / 2) * modifier for pred, modifier in zip(preds, error_increase)]

    return data


def add_tn_cat_conf_bounds(data: pd.DataFrame, tss_args):
    data['confidence'] = data['confidence'].astype(object)
    for idx, row in data.iterrows():
        data['confidence'].iloc[idx] = [row['confidence'] for _ in range(tss_args.horizon)]
    return data


class Differencer:
    def __init__(self):
        self.original_train_series = None
        self.diffed_train_series = None
        self.first_train_value = None
        self.last_train_value = None

    def diff(self, series: np.array) -> pd.Series:
        series = self._flatten_series(series)
        s = pd.Series(series)
        return s.shift(-1) - s

    def fit(self, series: np.array) -> None:
        series = self._flatten_series(series)
        self.first_train_value = series[0]
        self.last_train_value = series[-1]
        self.original_train_series = series
        self.diffed_train_series = self.diff(series)

    def transform(self, series: np.array) -> pd.Series:
        series = self._flatten_series(series)
        return self.diff(series).shift(1).fillna(0)

    def inverse_transform(self, series: pd.Series, init=None) -> pd.Series:
        origin = init if init else self.last_train_value
        s = pd.Series(origin)
        s = s.append(series).dropna()
        return s.expanding().sum()

    @staticmethod
    def _flatten_series(series: np.ndarray) -> np.ndarray:
        if len(series.shape) > 2:
            raise Exception(f"Input data should be shaped (A,) or (A, 1), got {series.shape}")
        elif len(series.shape) == 2:
            series = series.flatten()
        return series


def detect_freq_period(deltas: pd.DataFrame, tss) -> tuple:
    """
    Helper method that, based on the most popular interval for a time series, determines its seasonal peridiocity (sp).
    This bit of information can be crucial for good modelling with methods like ARIMA.

    Supported time intervals are:
        * 'year'
        * 'semestral'
        * 'quarter'
        * 'bimonthly'
        * 'monthly'
        * 'weekly'
        * 'daily'
        * 'hourly'
        * 'minute'
        * 'second'

    Note: all computations assume that the first provided `order_by` column is the one that specifies the sp.

    :param deltas: output of `get_delta`, has the most popular interval for each time series.
    :param tss: timeseries settings.

    :return: for all time series 1) a dictionary with its sp and 2) a dictionary with the detected sampling frequency
    """  # noqa
    secs_to_interval = {
        'yearly': 60 * 60 * 24 * 365,
        'quarterly': 60 * 60 * 24 * 365 // 4,
        'bimonthly': 60 * 60 * 24 * 31 * 2,
        'monthly': 60 * 60 * 24 * 31,
        'weekly': 60 * 60 * 24 * 7,
        'daily': 60 * 60 * 24,
        'hourly': 60 * 60,
        'minute': 60,
        'second': 1,
        'millisecond': 0.001,
        'microsecond': 1e-6,
        'nanosecond': 1e-9,
        'constant': 0
    }
    freq_to_period = {interval: period for (interval, period) in tss.interval_periods}
    for tag, period in (('yearly', 1), ('quarterly', 4), ('bimonthly', 6), ('monthly', 12),
                        ('weekly', 4), ('daily', 1), ('hourly', 24), ('minute', 1), ('second', 1), ('constant', 0)):
        if tag not in freq_to_period.keys():
            freq_to_period[tag] = period

    diffs = [(tag, abs(deltas - secs)) for tag, secs in secs_to_interval.items()]
    freq, min_diff = sorted(diffs, key=lambda x: x[1])[0]
    return freq_to_pandas(freq), freq_to_period.get(freq, 1)


def freq_to_pandas(freq, sample_row=None):
    mapping = {
        'constant': 'N',
        'nanosecond': 'N',
        'microsecond': 'us',
        'millisecond': 'ms',
        'second': 'S',
        'minute': 'T',
        'hourly': 'H',  # custom logic
        'daily': 'D',  # custom logic
        'weekly': 'W',  # anchor logic
        'monthly': 'M',  # custom logic
        'bimonthly': 'M',
        'quarterly': 'Q',  # anchor and custom logic
        'yearly': 'Y',  # anchor and custom logic
    }

    # TODO: implement custom dispatch for better precision, use row sample if available:
    #  pandas.pydata.org/docs/user_guide/timeseries.html
    return mapping[freq]


def max_pacf(data: pd.DataFrame, group_combinations, target, tss):
    def min_k(top_k, data):
        return min(top_k, len(data))

    top_k = 5
    k = min_k(top_k, data[target])
    candidate_sps = {'__default': (1 + np.argpartition(pacf(data[target].values)[1:], -k))[-k:].tolist()[::-1]}
    if tss.group_by:
        for group in group_combinations:
            if group != "__default":
                _, subset = get_group_matches(data, group, tss.group_by)
                try:
                    k = min_k(top_k, subset[target])
                    candidates = (1 + np.argpartition(pacf(subset[target].values)[1:], -k))[-k:].tolist()[::-1]
                    candidate_sps[group] = candidates
                except Exception:
                    candidate_sps[group] = None
            if not candidate_sps[group]:
                candidate_sps[group] = [1]

    return candidate_sps



import time
from copy import deepcopy
from typing import Dict, Optional, List

import numpy as np
import pandas as pd

import torch
from torch import nn
import torch_optimizer as ad_optim
from torch.cuda.amp import GradScaler
from torch.utils.data import DataLoader
from torch.optim.optimizer import Optimizer

from lightwood.api import dtype
from lightwood.api.types import PredictionArguments
from lightwood.encoder.base import BaseEncoder
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs
from lightwood.mixer.neural import Neural
from lightwood.mixer.helpers.ar_net import ArNet
from lightwood.mixer.helpers.default_net import DefaultNet
from lightwood.mixer.helpers.ts import _apply_stl_on_training, _stl_transform, _stl_inverse_transform
from lightwood.api.types import TimeseriesSettings


class NeuralTs(Neural):
    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            timeseries_settings: TimeseriesSettings,
            target_encoder: BaseEncoder,
            net: str,
            fit_on_dev: bool,
            search_hyperparameters: bool,
            ts_analysis: Dict[str, Dict],
            n_epochs: Optional[int] = None,
            use_stl: Optional[bool] = False
    ):
        """
        Subclassed Neural mixer used for time series forecasting scenarios. 
        
        :param stop_after: How long the total fitting process should take
        :param target: Name of the target column
        :param dtype_dict: Data type dictionary
        :param timeseries_settings: TimeseriesSettings object for time-series tasks, refer to its documentation for available settings.
        :param target_encoder: Reference to the encoder used for the target
        :param net: The network type to use (`DeafultNet` or `ArNet`)
        :param fit_on_dev: If we should fit on the dev dataset
        :param search_hyperparameters: If the network should run a more through hyperparameter search (currently disabled)
        :param n_epochs: amount of epochs that the network will be trained for. Supersedes all other early stopping criteria if specified.
        """ # noqa
        super().__init__(
            stop_after,
            target,
            dtype_dict,
            target_encoder,
            net,
            fit_on_dev,
            search_hyperparameters,
            n_epochs,
        )
        self.timeseries_settings = timeseries_settings
        assert self.timeseries_settings.is_timeseries
        self.ts_analysis = ts_analysis
        self.net_class = DefaultNet if net == 'DefaultNet' else ArNet
        self.stable = True
        self.use_stl = use_stl

    def _select_criterion(self) -> torch.nn.Module:
        if self.dtype_dict[self.target] in (dtype.integer, dtype.float, dtype.num_tsarray, dtype.quantity):
            criterion = nn.L1Loss()
        else:
            criterion = super()._select_criterion()

        return criterion

    def _select_optimizer(self) -> Optimizer:
        optimizer = ad_optim.Ranger(self.model.parameters(), lr=self.lr)
        return optimizer

    def _fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        :param train_data: The network is fit/trained on this
        :param dev_data: Data used for early stopping and hyperparameter determination
        """  # noqa
        self.started = time.time()
        original_train = deepcopy(train_data.data_frame)
        original_dev = deepcopy(dev_data.data_frame)

        # Use STL blocks if available
        if self.use_stl and self.ts_analysis.get('stl_transforms', False):
            _apply_stl_on_training(train_data, dev_data, self.target, self.timeseries_settings, self.ts_analysis)

        # ConcatedEncodedDs
        self.batch_size = min(200, int(len(train_data) / 10))
        self.batch_size = max(40, self.batch_size)

        dev_dl = DataLoader(dev_data, batch_size=self.batch_size, shuffle=False)
        train_dl = DataLoader(train_data, batch_size=self.batch_size, shuffle=False)

        self.lr = 1e-4
        self.num_hidden = 1

        # Find learning rate
        # keep the weights
        self._init_net(train_data)
        self.lr, self.model = self._find_lr(train_dl)

        # Keep on training
        optimizer = self._select_optimizer()
        criterion = self._select_criterion()
        scaler = GradScaler()

        # Only 0.8 of the remaining time budget is used to allow some time for the final tuning and partial fit
        self.model, epoch_to_best_model, _ = self._max_fit(
            train_dl, dev_dl, criterion, optimizer, scaler, (self.stop_after - (time.time() - self.started)) * 0.8,
            return_model_after=20000)

        self.epochs_to_best += epoch_to_best_model

        # restore dfs
        train_data.data_frame = original_train
        dev_data.data_frame = original_dev

        if self.fit_on_dev:
            self.partial_fit(dev_data, train_data)

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        self._fit(train_data, dev_data)

    def __call__(self, ds: EncodedDs,
                 args: PredictionArguments = PredictionArguments()
                 ) -> pd.DataFrame:
        original_df = deepcopy(ds.data_frame)

        self.model = self.model.eval()
        decoded_predictions = []
        all_probs: List[List[float]] = []
        rev_map = {}

        length = sum(ds.encoded_ds_lenghts) if isinstance(ds, ConcatedEncodedDs) else len(ds)
        pred_cols = [f'prediction_{i}' for i in range(self.timeseries_settings.horizon)]
        ydf = pd.DataFrame(0,  # zero-filled
                           index=np.arange(length),
                           dtype=object,
                           columns=pred_cols)

        if self.use_stl and self.ts_analysis.get('stl_transforms', False):
            ds.data_frame = _stl_transform(ydf, ds, self.target, self.timeseries_settings, self.ts_analysis)

        with torch.no_grad():
            for idx, (X, Y) in enumerate(ds):
                X = X.to(self.model.device)
                Yh = self.model(X)
                Yh = torch.unsqueeze(Yh, 0) if len(Yh.shape) < 2 else Yh

                kwargs = {}
                for dep in self.target_encoder.dependencies:
                    kwargs['dependency_data'] = {dep: ds.data_frame.iloc[idx][[dep]].values}

                if args.predict_proba and self.supports_proba:
                    decoded_prediction, probs, rev_map = self.target_encoder.decode_probabilities(Yh, **kwargs)
                    all_probs.append(probs)
                else:
                    decoded_prediction = self.target_encoder.decode(Yh, **kwargs)

                decoded_predictions.extend(decoded_prediction)

        decoded_predictions = np.array(decoded_predictions)
        if len(decoded_predictions.shape) == 1:
            decoded_predictions = np.expand_dims(decoded_predictions, axis=1)
        ydf[pred_cols] = decoded_predictions

        if self.use_stl and self.ts_analysis.get('stl_transforms', False):
            ydf = _stl_inverse_transform(ydf, ds, self.timeseries_settings, self.ts_analysis)

        ydf['prediction'] = ydf.values.tolist()

        if self.timeseries_settings.horizon == 1:
            ydf['prediction'] = [p[0] for p in ydf['prediction']]

        if args.predict_proba and self.supports_proba:
            raw_predictions = np.array(all_probs).squeeze(axis=1)

            for idx, label in enumerate(rev_map.values()):
                ydf[f'__mdb_proba_{label}'] = raw_predictions[:, idx]

        # TODO: make this part of the base mixer class? to avoid repetitive code
        #  and ensure other contribs don't accidentally modify the df
        ds.data_frame = original_df
        return ydf[['prediction']]
